@@ -60,9 +60,19 @@
 #define max(a,b)			(a > b ? a : b)
 #endif
 
+#if !defined(__has_attribute)
+#define __has_attribute(x) 0
+#endif
+
+#if __has_attribute(used) || defined(__GNUC__)
+#define ATTRIBUTE_USED __attribute__((used))
+#else
+#define ATTRIBUTE_USED
+#endif
+
 /* avoid direct floating point comparison */
-#define gdip_near_zero(value)		((value >= -0.0001f) && (value <= 0.0001f))
-#define gdip_near_one(value)		((value >= 0.9999f) && (value <= 1.0001f))
+#define gdip_near_zero(value)		((value >= -0.00059604645f) && (value <= 0.00059604645f))
+#define gdip_near_one(value)		gdip_near_zero (value - 1.0f)
 
 /* avoid integer overflows when int/float/int conversion are used (see #79643) */
 #define SAFE_FLOAT_TO_UINT32(value)    ((value < 0) ? 0 : (value >= G_MAXINT32) ? G_MAXINT32 : (((int)value) < 0 ? G_MAXINT32 : (int)value))
@@ -81,6 +91,7 @@
 
 /* avoid floating point division/multiplications when pre-multiplying the alpha channel with R, G and B values */
 extern const BYTE pre_multiplied_table[256][256];
+extern BOOL gdiplusInitialized;
 
 #if CAIRO_VERSION < CAIRO_VERSION_ENCODE(1,6,0)
 	/* older cairo has a (signed) 15(1)/16(2)bits pixel positioning, while GDI+ use (signed) 23 bits (infinity).
@@ -104,18 +115,32 @@ float gdip_get_display_dpi () GDIP_INTERNAL;
 GpStatus gdip_get_status (cairo_status_t status) GDIP_INTERNAL;
 GpStatus gdip_get_pattern_status (cairo_pattern_t *pat) GDIP_INTERNAL;
 
-gchar *ucs2_to_utf8 (const gunichar2 *ucs2, int length) GDIP_INTERNAL;
+gchar *utf16_to_utf8 (const gunichar2 *ucs2, int length) GDIP_INTERNAL;
 BOOL utf8_to_ucs2 (const gchar *utf8, gunichar2 *ucs2, int ucs2_len) GDIP_INTERNAL;
 int utf8_encode_ucs2char (gunichar2 unichar, unsigned char *dest) GDIP_INTERNAL;
 
 /* for drawing curves */
 GpPointF *convert_points (const GpPoint *points, int count) GDIP_INTERNAL;
+GpRectF *convert_rects (const GpRect *rect, int count) GDIP_INTERNAL;
 GpPointF *gdip_closed_curve_tangents (int terms, const GpPointF *points, int count, float tension) GDIP_INTERNAL;
 GpPointF *gdip_open_curve_tangents (int terms,  const GpPointF *points, int count, float tension) GDIP_INTERNAL;
 
+/* Conversion functions. */
+void gdip_RectF_from_Rect (const GpRect* rect, GpRectF* rectf) GDIP_INTERNAL;
+void gdip_Rect_from_RectF (const GpRectF* rectf, GpRect* rect) GDIP_INTERNAL;
+void gdip_PointF_from_Point (const GpPoint* point, GpPointF* pointf) GDIP_INTERNAL;
+void gdip_Point_from_PointF (const GpPointF* pointf, GpPoint* point) GDIP_INTERNAL;
+
+/* Rect functions */
+void gdip_normalize_rectangle (const GpRectF *rect, GpRectF *normalized) GDIP_INTERNAL;
+BOOL gdip_is_rectF_empty (const GpRectF *rect, BOOL allowNegative) GDIP_INTERNAL;
+
 /* memory */
 void* gdip_realloc (void *org, int size) GDIP_INTERNAL;
-void* gdip_calloc (size_t nelem, size_t elsize) GDIP_INTERNAL; 
+void* gdip_calloc (size_t nelem, size_t elsize) GDIP_INTERNAL;
+
+/* checksums */
+DWORD gdip_crc32 (const BYTE *buf, size_t size) GDIP_INTERNAL;
 
 #include "general.h"
 
