@@ -32,6 +32,7 @@
 #include "gdiplus-private.h"
 #include "image-private.h"
 #include "imageattributes-private.h"
+#include "general-private.h"
 #include "graphics-private.h"
 #include "matrix.h"
 
@@ -62,8 +63,15 @@ GUID gdip_membmp_image_format_guid = {0xb96b3caaU, 0x0728U, 0x11d3U, {0x9d, 0x7b
 /*
  * encoder param guids
  */
-GUID GdipEncoderQuality = {0x1d5be4b5U, 0xfa4aU, 0x452dU, {0x9c, 0xdd, 0x5d, 0xb3, 0x51, 0x05, 0xe7, 0xeb}};
-GUID GdipEncoderCompression = {0xe09d739dU, 0xccd4U, 0x44eeU, {0x8e, 0xba, 0x3f, 0xbf, 0x8b, 0xe4, 0xfc, 0x58}};
+GUID GdipEncoderCompression = {0x0E09D739DU, 0x0CCD4U, 0x44EEU, {0x8E, 0x0BA, 0x3F, 0x0BF, 0x8B, 0x0E4, 0x0FC, 0x58}};
+GUID GdipEncoderColorDepth = {0x66087055U, 0x0AD66U, 0x4C7CU, {0x9A, 0x18, 0x38, 0x0A2, 0x31, 0x0B, 0x83, 0x37}};
+GUID GdipEncoderSaveFlag = {0x292266FCU, 0x0AC40U, 0x47BFU, {0x8C, 0x0FC, 0x0A8, 0x5B, 0x89, 0x0A6, 0x55, 0x0DE}};
+GUID GdipEncoderSaveAsCMYK = {0x0A219BBC9U, 0x0A9DU, 0x4005U, {0x0A3, 0x0EE, 0x3A, 0x42, 0x1B, 0x8B, 0x0B0, 0x6C}};
+GUID GdipEncoderImageItems = {0x63875E13U, 0x1F1DU, 0x45ABU, {0x91, 0x95, 0x0A2, 0x9B, 0x60, 0x66, 0x0A6, 0x50}};
+GUID GdipEncoderTransformation = {0x8D0EB2D1U, 0x0A58EU, 0x4EA8U, {0x0AA, 0x14, 0x10, 0x80, 0x74, 0x0B7, 0x0B6, 0x0F9}};
+GUID GdipEncoderQuality = {0x1D5BE4B5U, 0x0FA4AU, 0x452DU, {0x9C, 0x0DD, 0x5D, 0x0B3, 0x51, 0x5, 0x0E7, 0x0EB}};
+GUID GdipEncoderLuminanceTable = {0x0EDB33BCEU, 0x266U, 0x4A77U, {0x0B9, 0x4, 0x27, 0x21, 0x60, 0x99, 0x0E7, 0x17}};
+GUID GdipEncoderChrominanceTable = {0x0F2E455DCU, 0x9B3U, 0x4316U, {0x82, 0x60, 0x67, 0x6A, 0x0DA, 0x32, 0x48, 0x1C}};
 
 #define DECODERS_SUPPORTED 8
 #define ENCODERS_SUPPORTED 5
@@ -107,6 +115,9 @@ static const int nonplaceable_wmf_sig_size = 6;
 static BOOL
 signature_match (char *data, size_t data_size, int size, int count, BYTE* sig_pattern, BYTE* sig_mask)
 {
+	if (data_size < size)
+		return FALSE;
+
 	int sig;
 	for (sig = 0; sig < (size * count); sig += size) {
 		BOOL match = TRUE;
@@ -153,8 +164,6 @@ get_image_format (char *sig_read, size_t size_read, ImageFormat *final)
 static cairo_filter_t
 gdip_get_cairo_filter (InterpolationMode imode)
 {
-	cairo_filter_t filter;
-
 	switch (imode) {
 	case InterpolationModeHighQuality:
 	case InterpolationModeHighQualityBilinear:
@@ -203,11 +212,11 @@ gdip_flip_x (GpImage *image)
 	height = image->active_bitmap->height;
 	pixel_size = gdip_get_pixel_format_components (image->active_bitmap->pixel_format) * gdip_get_pixel_format_depth (image->active_bitmap->pixel_format) / 8;
 	line = GdipAlloc (stride);
-	src = (BYTE *) image->active_bitmap->scan0;
-
-	if (line == NULL) {
+	if (!line) {
 		return OutOfMemory;
 	}
+
+	src = (BYTE *) image->active_bitmap->scan0;
 	
 	for (i = 0; i < height; i++, src += stride) {
 		memcpy (line, src, stride);	/* Save original line */
@@ -235,13 +244,13 @@ gdip_flip_y (GpImage *image)
 	stride = image->active_bitmap->stride;
 	height = image->active_bitmap->height;
 	line = GdipAlloc (stride);
+	if (!line) {
+		return OutOfMemory;
+	}
+
 	src = (BYTE *) image->active_bitmap->scan0;
 	trg = (BYTE *) image->active_bitmap->scan0;
 	trg +=  (height-1) * stride;
-
-	if (line == NULL) {
-		return OutOfMemory;
-	}
 	
 	for (i = 0; i < (height /2); i++, src += stride, trg -= stride) {			
 		memcpy (line, trg, stride);	/* Save target line*/
@@ -266,7 +275,7 @@ GdipDisposeImage (GpImage *image)
 	case ImageTypeMetafile:
 		return gdip_metafile_dispose ((GpMetafile*)image);
 	default: /* imageUndefined */
-		g_warning ("unknown image type couldn't be disposed, ptr = %d, type %d", image, image->type);
+		g_warning ("unknown image type couldn't be disposed, ptr = %p, type %d", (void*)image, image->type);
 		return Ok;
 	}
 }
@@ -313,8 +322,15 @@ GdipGetImageGraphicsContext (GpImage *image, GpGraphics **graphics)
 				image->active_bitmap->width, image->active_bitmap->height, image->active_bitmap->stride);
 
 	gfx = gdip_graphics_new (surface);
+	if (!gfx)
+		return OutOfMemory;
+
 	gfx->dpi_x = image->active_bitmap->dpi_horz <= 0 ? gdip_get_display_dpi () : image->active_bitmap->dpi_horz;
 	gfx->dpi_y = image->active_bitmap->dpi_vert <= 0 ? gdip_get_display_dpi () : image->active_bitmap->dpi_vert;
+	gfx->bounds.Width = image->active_bitmap->width;
+	gfx->bounds.Height = image->active_bitmap->height;
+	gfx->orig_bounds.Width = image->active_bitmap->width;
+	gfx->orig_bounds.Height = image->active_bitmap->height;
 	cairo_surface_destroy (surface);
 
 	gfx->image = image;
@@ -329,32 +345,22 @@ GdipGetImageGraphicsContext (GpImage *image, GpGraphics **graphics)
 GpStatus WINGDIPAPI 
 GdipDrawImageI (GpGraphics *graphics, GpImage *image, INT x, INT y)
 {
-	if (!image)
-		return InvalidParameter;
-
-	switch (image->type) {
-	case ImageTypeBitmap: {
-		BitmapData *data = image->active_bitmap;
-		return GdipDrawImageRect (graphics, image, x, y, data->width, data->height);
-	}
-	case ImageTypeMetafile: {
-		MetafileHeader *metaheader = &((GpMetafile*)image)->metafile_header;
-		return GdipDrawImageRect (graphics, image, x, y, metaheader->Width, metaheader->Height);
-	}
-	default:
-		return InvalidParameter;
-	}
+	return GdipDrawImage (graphics, image, x, y);
 }
 
 GpStatus WINGDIPAPI 
 GdipDrawImage (GpGraphics *graphics, GpImage *image, REAL x, REAL y)
 {
+	if (!graphics)
+		return InvalidParameter;
+	if (graphics->state == GraphicsStateBusy)
+		return ObjectBusy;
 	if (!image)
 		return InvalidParameter;
 
 	switch (image->type) {
 	case ImageTypeBitmap: {
-		BitmapData *data = image->active_bitmap;
+		ActiveBitmapData *data = image->active_bitmap;
 		return GdipDrawImageRect (graphics, image, x, y, data->width, data->height);
 	}
 	case ImageTypeMetafile: {
@@ -384,11 +390,12 @@ GdipDrawImageRect (GpGraphics *graphics, GpImage *image, REAL x, REAL y, REAL wi
 	BYTE *premul = NULL;
 	cairo_surface_t *original = NULL;
 
-	if (!graphics || !image)
+	if (!graphics)
+		return InvalidParameter;
+	if (graphics->state == GraphicsStateBusy)
+		return ObjectBusy;
+	if (!image)
 		return InvalidParameter;
-
-	if ((width <= 0) || (height <= 0))
-		return Ok;
 			
 	if (image->type == ImageTypeBitmap) {
 		/* check does not apply to metafiles, and it's better be done before converting the image */
@@ -407,6 +414,9 @@ GdipDrawImageRect (GpGraphics *graphics, GpImage *image, REAL x, REAL y, REAL wi
 		}
 	}
 
+	if ((width == 0) || (height == 0))
+		return Ok;
+
 	/* conversion must be done after the recursive call to GdipDrawImageRect to remove the indexed bitmap */
 	if (!OPTIMIZE_CONVERSION (graphics)) {
 		x = gdip_unitx_convgr (graphics, x);
@@ -423,7 +433,6 @@ GdipDrawImageRect (GpGraphics *graphics, GpImage *image, REAL x, REAL y, REAL wi
 
 		metacontext = gdip_metafile_play_setup ((GpMetafile*)image, graphics, x, y, width, height);
 
-		cairo_translate (graphics->ct, x, y);
 		status = gdip_metafile_play (metacontext);
 
 		gdip_metafile_play_cleanup (metacontext);
@@ -433,11 +442,10 @@ GdipDrawImageRect (GpGraphics *graphics, GpImage *image, REAL x, REAL y, REAL wi
 	/* Create a surface for this bitmap if one doesn't exist */
 	gdip_bitmap_ensure_surface (image);
 	
-	if (graphics->type != gtMemoryBitmap &&
-		gdip_bitmap_format_needs_premultiplication (image)) {
+	if (gdip_bitmap_format_needs_premultiplication (image)) {
 		premul = gdip_bitmap_get_premultiplied_scan0 (image);
 		if (premul) {
-			BitmapData *data = image->active_bitmap;
+			ActiveBitmapData *data = image->active_bitmap;
 			original = cairo_image_surface_create_for_data (premul, CAIRO_FORMAT_ARGB32,
 															data->width, data->height, data->stride);
 		}
@@ -496,9 +504,15 @@ GdipDrawImagePoints (GpGraphics *graphics, GpImage *image, GDIPCONST GpPointF *d
 	MetafilePlayContext *metacontext = NULL;
 	BYTE *premul = NULL;
 	cairo_surface_t *original = NULL;
-	
-	if (!graphics || !image || !dstPoints || (count != 3))
+
+	if (!graphics || !dstPoints || count <= 0)
 		return InvalidParameter;
+	if (graphics->state == GraphicsStateBusy)
+		return ObjectBusy;
+	if (!image || (count != 3 && count != 4))
+		return InvalidParameter;
+	if (count == 4)
+		return NotImplemented;
 
 	cairo_new_path (graphics->ct);
 
@@ -530,7 +544,7 @@ GdipDrawImagePoints (GpGraphics *graphics, GpImage *image, GDIPCONST GpPointF *d
 		metacontext = gdip_metafile_play_setup ((GpMetafile*)image, graphics, tRect.X, tRect.Y, tRect.Width, tRect.Height);
 
 		cairo_get_matrix (graphics->ct, &orig_matrix);
-		cairo_set_matrix (graphics->ct, matrix);
+		gdip_cairo_set_matrix (graphics, matrix);
 		status = gdip_metafile_play (metacontext);
 		GdipDeleteMatrix (matrix);
 
@@ -541,11 +555,10 @@ GdipDrawImagePoints (GpGraphics *graphics, GpImage *image, GDIPCONST GpPointF *d
 	/* Create a surface for this bitmap if one doesn't exist */
 	gdip_bitmap_ensure_surface (image);
 
-	if (graphics->type != gtMemoryBitmap &&
-		gdip_bitmap_format_needs_premultiplication (image)) {
+	if (gdip_bitmap_format_needs_premultiplication (image)) {
 		premul = gdip_bitmap_get_premultiplied_scan0 (image);
 		if (premul) {
-			BitmapData *data = image->active_bitmap;
+			ActiveBitmapData *data = image->active_bitmap;
 			original = cairo_image_surface_create_for_data (premul, CAIRO_FORMAT_ARGB32,
 															data->width, data->height, data->stride);
 		}
@@ -562,7 +575,7 @@ GdipDrawImagePoints (GpGraphics *graphics, GpImage *image, GDIPCONST GpPointF *d
 	cairo_pattern_reference(org_pattern);
 	
 	cairo_get_matrix(graphics->ct, &orig_matrix);
-	cairo_set_matrix (graphics->ct, matrix);
+	gdip_cairo_set_matrix (graphics, matrix);
 	cairo_set_source_surface (graphics->ct, image->surface, 0, 0);
 	
 	cairo_paint (graphics->ct);	
@@ -584,18 +597,20 @@ GdipDrawImagePoints (GpGraphics *graphics, GpImage *image, GDIPCONST GpPointF *d
 GpStatus WINGDIPAPI
 GdipDrawImagePointsI (GpGraphics *graphics, GpImage *image, GDIPCONST GpPoint *dstPoints, INT count)
 {
-	GpPointF points[3];
-	int i;
-	
-	if (!dstPoints || (count != 3))
+	GpStatus status;
+	GpPointF *pointsF;
+
+	if (!dstPoints || count < 0)
 		return InvalidParameter;
 
-	for (i = 0; i < 3; i++) {
-		points[i].X = dstPoints[i].X;
-		points[i].Y = dstPoints[i].Y;
-	}
-	
-	return GdipDrawImagePoints (graphics, image, points, 3);
+	pointsF = convert_points (dstPoints, count);
+	if (!pointsF)
+		return OutOfMemory;
+
+	status = GdipDrawImagePoints (graphics, image, pointsF, count);
+
+	GdipFree (pointsF);
+	return status;
 }
 
 GpStatus WINGDIPAPI
@@ -621,6 +636,7 @@ GdipDrawImageRectRect (GpGraphics *graphics, GpImage *image,
                        GDIPCONST GpImageAttributes *imageAttributes,
                        DrawImageAbort callback, void *callbackData)
 {
+	GpStatus status;
 	cairo_pattern_t	*pattern;
 	cairo_pattern_t	*orig;
 	cairo_matrix_t	mat;
@@ -631,7 +647,11 @@ GdipDrawImageRectRect (GpGraphics *graphics, GpImage *image,
 	BYTE			*premul = NULL;
 	cairo_surface_t	*original = NULL;
 	
-	if (!graphics || !image)
+	if (!graphics)
+		return InvalidParameter;
+	if (graphics->state == GraphicsStateBusy)
+		return ObjectBusy;
+	if (!image)
 		return InvalidParameter;
 
 	switch (srcUnit) {
@@ -652,15 +672,15 @@ GdipDrawImageRectRect (GpGraphics *graphics, GpImage *image,
 
 	if (image->type == ImageTypeBitmap) {
 		if (gdip_is_an_indexed_pixelformat (image->active_bitmap->pixel_format)) {
-			GpStatus status = OutOfMemory;
 			GpBitmap *rgb_bitmap = gdip_convert_indexed_to_rgb (image);
-			if (rgb_bitmap) {
-				status = GdipDrawImageRectRect (graphics, rgb_bitmap,
-					dstx, dsty, dstwidth, dstheight,
-					srcx, srcy, srcwidth, srcheight,
-					srcUnit, imageAttributes, callback, callbackData);
-				GdipDisposeImage (rgb_bitmap);
-			}
+			if (!rgb_bitmap)
+				return OutOfMemory;
+
+			status = GdipDrawImageRectRect (graphics, rgb_bitmap,
+				dstx, dsty, dstwidth, dstheight,
+				srcx, srcy, srcwidth, srcheight,
+				srcUnit, imageAttributes, callback, callbackData);
+			GdipDisposeImage (rgb_bitmap);
 
 			return status;
 		}
@@ -670,7 +690,7 @@ GdipDrawImageRectRect (GpGraphics *graphics, GpImage *image,
 	}
 
 	/* see OPTIMIZE_CONVERSION in general.h */
-	if ((srcUnit != UnitPixel) && (srcUnit != UnitWorld) && ((srcUnit != UnitDisplay) || (graphics->type != gtPostScript))) {
+	if (srcUnit != UnitPixel && graphics->type != gtPostScript) {
 		dstx = gdip_unit_conversion (srcUnit, UnitCairoPoint, graphics->dpi_x, graphics->type, dstx);
 		dsty = gdip_unit_conversion (srcUnit, UnitCairoPoint, graphics->dpi_y, graphics->type, dsty);
 		dstwidth = gdip_unit_conversion (srcUnit, UnitCairoPoint, graphics->dpi_x, graphics->type, dstwidth);
@@ -683,7 +703,9 @@ GdipDrawImageRectRect (GpGraphics *graphics, GpImage *image,
 
 	org = dest = image->active_bitmap->scan0; 
 	org_format = image->active_bitmap->pixel_format;
-	gdip_process_bitmap_attributes (image, &dest, (GpImageAttributes *) imageAttributes, &allocated);
+	status = gdip_process_bitmap_attributes (image, &dest, (GpImageAttributes *) imageAttributes, &allocated);
+	if (status != Ok)
+		return status;
 
 	/*  If allocated is true we have a newly allocated and altered Scan0 in dest */
 	if (allocated) {
@@ -728,15 +750,22 @@ GdipDrawImageRectRect (GpGraphics *graphics, GpImage *image,
 		if (flipXOn) {			
 			/* We're ok just cloning the bitmap, we don't need the image data
 			 * and we destroy it before we leave this function */
-			gdip_bitmap_clone (image, &imgflipX);
-			gdip_flip_x (imgflipX);	
+			status = gdip_bitmap_clone (image, &imgflipX);
+			if (status != Ok)
+				return status;
+
+			status = gdip_flip_x (imgflipX);
+			if (status != Ok) {
+				gdip_bitmap_dispose (imgflipX);
+				return status;
+			}
+
 			gdip_bitmap_ensure_surface (imgflipX);			
 
-			if (graphics->type != gtMemoryBitmap &&
-				gdip_bitmap_format_needs_premultiplication (imgflipX)) {
+			if (gdip_bitmap_format_needs_premultiplication (imgflipX)) {
 				premulX = gdip_bitmap_get_premultiplied_scan0 (imgflipX);
 				if (premulX) {
-					BitmapData *data = imgflipX->active_bitmap;
+					ActiveBitmapData *data = imgflipX->active_bitmap;
 					originalX = cairo_image_surface_create_for_data (premulX, CAIRO_FORMAT_ARGB32, data->width, data->height, data->stride);
 				}
 			}
@@ -747,15 +776,22 @@ GdipDrawImageRectRect (GpGraphics *graphics, GpImage *image,
 		}
 		
 		if (flipYOn) {			
-			gdip_bitmap_clone (image, &imgflipY);
-			gdip_flip_y (imgflipY);	
+			status = gdip_bitmap_clone (image, &imgflipY);
+			if (status != Ok)
+				return status;
+
+			status = gdip_flip_y (imgflipY);
+			if (status != Ok) {
+				gdip_bitmap_dispose (imgflipY);
+				return status;
+			}
+
 			gdip_bitmap_ensure_surface (imgflipY);			
 			
-			if (graphics->type != gtMemoryBitmap &&
-				gdip_bitmap_format_needs_premultiplication (imgflipY)) {
+			if (gdip_bitmap_format_needs_premultiplication (imgflipY)) {
 				premulY = gdip_bitmap_get_premultiplied_scan0 (imgflipY);
 				if (premulY) {
-					BitmapData *data = imgflipY->active_bitmap;
+					ActiveBitmapData *data = imgflipY->active_bitmap;
 					originalY = cairo_image_surface_create_for_data (premulY, CAIRO_FORMAT_ARGB32, data->width, data->height, data->stride);
 				}
 			}
@@ -766,16 +802,28 @@ GdipDrawImageRectRect (GpGraphics *graphics, GpImage *image,
 		}
 		
 		if (flipXOn && flipYOn) {			
-			gdip_bitmap_clone (image, &imgflipXY);
-			gdip_flip_x (imgflipXY);	
-			gdip_flip_y (imgflipXY);	
+			status = gdip_bitmap_clone (image, &imgflipXY);
+			if (status != Ok)
+				return status;
+
+			status = gdip_flip_x (imgflipXY);
+			if (status != Ok) {
+				gdip_bitmap_dispose (imgflipXY);
+				return status;
+			}
+
+			status = gdip_flip_y (imgflipXY);
+			if (status != Ok) {
+				gdip_bitmap_dispose (imgflipXY);
+				return Ok;
+			}
+
 			gdip_bitmap_ensure_surface (imgflipXY);			
 
-			if (graphics->type != gtMemoryBitmap &&
-				gdip_bitmap_format_needs_premultiplication (imgflipXY)) {
+			if (gdip_bitmap_format_needs_premultiplication (imgflipXY)) {
 				premulXY = gdip_bitmap_get_premultiplied_scan0 (imgflipXY);
 				if (premulXY) {
-					BitmapData *data = imgflipXY->active_bitmap;
+					ActiveBitmapData *data = imgflipXY->active_bitmap;
 					originalXY = cairo_image_surface_create_for_data (premulXY, CAIRO_FORMAT_ARGB32, data->width, data->height, data->stride);
 				}
 			}
@@ -787,11 +835,10 @@ GdipDrawImageRectRect (GpGraphics *graphics, GpImage *image,
 		
 		gdip_bitmap_ensure_surface (image);
 
-		if (graphics->type != gtMemoryBitmap &&
-			gdip_bitmap_format_needs_premultiplication (image)) {
+		if (gdip_bitmap_format_needs_premultiplication (image)) {
 			premul = gdip_bitmap_get_premultiplied_scan0 (image);
 			if (premul) {
-				BitmapData *data = image->active_bitmap;
+				ActiveBitmapData *data = image->active_bitmap;
 				original = cairo_image_surface_create_for_data (premul, CAIRO_FORMAT_ARGB32, data->width, data->height, data->stride);
 			}
 		}
@@ -883,11 +930,10 @@ GdipDrawImageRectRect (GpGraphics *graphics, GpImage *image,
 
 		gdip_bitmap_ensure_surface (image);
 
-		if (graphics->type != gtMemoryBitmap &&
-			gdip_bitmap_format_needs_premultiplication (image)) {
+		if (gdip_bitmap_format_needs_premultiplication (image)) {
 			premul = gdip_bitmap_get_premultiplied_scan0 (image);
 			if (premul) {
-				BitmapData *data = image->active_bitmap;
+				ActiveBitmapData *data = image->active_bitmap;
 				original = cairo_image_surface_create_for_data (premul, CAIRO_FORMAT_ARGB32, data->width, data->height, data->stride);
 			}
 		}
@@ -966,9 +1012,30 @@ GdipDrawImagePointsRect (GpGraphics *graphics, GpImage *image, GDIPCONST GpPoint
 	GpMatrix *matrix = NULL;
 	cairo_matrix_t orig_matrix;
 
-	if (!graphics || !image || !points || (count < 3))
+	if (!graphics || !points || count <= 0)
 		return InvalidParameter;
-	if (count > 3)
+	if (graphics->state == GraphicsStateBusy)
+		return ObjectBusy;
+	if (!image || (count != 3 && count != 4))
+		return InvalidParameter;
+
+	switch (srcUnit) {
+	case UnitPixel:
+		break;
+	case UnitPoint:
+	case UnitInch:
+	case UnitDocument:
+	case UnitMillimeter:
+		if (graphics->type != gtPostScript)
+			return NotImplemented; /* GDI+ returns the same */
+		break;
+	case UnitWorld:
+	case UnitDisplay:
+	default:
+		return InvalidParameter;
+	}
+
+	if (count == 4)
 		return NotImplemented;
 
 	rect.X = 0; 
@@ -983,15 +1050,18 @@ GdipDrawImagePointsRect (GpGraphics *graphics, GpImage *image, GDIPCONST GpPoint
 	}
 
 	status = GdipCreateMatrix3 (&rect, points, &matrix);
-	if (status == Ok) {
-		cairo_get_matrix (graphics->ct, &orig_matrix);
-		cairo_set_matrix (graphics->ct, matrix);
-		status = GdipDrawImageRectRect (graphics, image, rect.X, rect.Y, rect.Width, rect.Height, srcx, srcy, 
-			srcwidth, srcheight, srcUnit, imageAttributes, callback, callbackData);
-		cairo_set_matrix (graphics->ct, &orig_matrix);
-	}
-	if (matrix)
+	if (status != Ok) {
 		GdipDeleteMatrix (matrix);
+		return status;
+	}
+
+	cairo_get_matrix (graphics->ct, &orig_matrix);
+	gdip_cairo_set_matrix (graphics, matrix);
+	status = GdipDrawImageRectRect (graphics, image, rect.X, rect.Y, rect.Width, rect.Height, srcx, srcy, 
+		srcwidth, srcheight, srcUnit, imageAttributes, callback, callbackData);
+	cairo_set_matrix (graphics->ct, &orig_matrix);
+	
+	GdipDeleteMatrix (matrix);
 	return status;
 }
 
@@ -1000,21 +1070,16 @@ GdipDrawImagePointsRectI (GpGraphics *graphics, GpImage *image, GDIPCONST GpPoin
 	INT srcwidth, INT srcheight, GpUnit srcUnit, GDIPCONST GpImageAttributes *imageAttributes, DrawImageAbort callback,
 	void *callbackData)
 {
-	GpPointF pf[3];
+	GpPointF *pointsF;
 
-	if (!points || (count < 3))
+	if (!points || count < 0)
 		return InvalidParameter;
-	if (count > 3)
-		return NotImplemented;
 
-	pf[0].X = points[0].X;
-	pf[0].Y = points[0].Y;
-	pf[1].X = points[1].X;
-	pf[1].Y = points[1].Y;
-	pf[2].X = points[2].X;
-	pf[2].Y = points[2].Y;
+	pointsF = convert_points (points, count);
+	if (!pointsF)
+		return OutOfMemory;
 
-	return GdipDrawImagePointsRect (graphics, image, (GDIPCONST GpPointF*)&pf, count, srcx, srcy, srcwidth, srcheight, 
+	return GdipDrawImagePointsRect (graphics, image, pointsF, count, srcx, srcy, srcwidth, srcheight, 
 		srcUnit, imageAttributes, callback, callbackData);
 }
 
@@ -1024,12 +1089,27 @@ GdipDrawImagePointsRectI (GpGraphics *graphics, GpImage *image, GDIPCONST GpPoin
 GpStatus WINGDIPAPI
 GdipLoadImageFromStream (void *stream, GpImage **image)
 {
+	if (!gdiplusInitialized)
+		return GdiplusNotInitialized;
+
+	if (!stream || !image)
+		return InvalidParameter;
+
 	return NotImplemented; /* GdipLoadImageFromStream - not supported */
 }
 
+GpStatus WINGDIPAPI
+GdipLoadImageFromStreamICM (void *stream, GpImage **image)
+{
+	return GdipLoadImageFromStream (stream, image);	
+}
+
 GpStatus WINGDIPAPI
 GdipSaveImageToStream (GpImage *image, void *stream, GDIPCONST CLSID *encoderCLSID, GDIPCONST EncoderParameters *params)
 {
+	if (!stream || !image)
+		return InvalidParameter;
+
 	return NotImplemented; /* GdipSaveImageToStream - not supported */
 }
 
@@ -1044,11 +1124,14 @@ GdipLoadImageFromFile (GDIPCONST WCHAR *file, GpImage **image)
 	char		*file_name = NULL;
 	char		format_peek[MAX_CODEC_SIG_LENGTH];
 	int		format_peek_sz;
+
+	if (!gdiplusInitialized)
+		return GdiplusNotInitialized;
 	
 	if (!image || !file)
 		return InvalidParameter;
 	
-	file_name = (char *) ucs2_to_utf8 ((const gunichar2 *)file, -1);
+	file_name = (char *) utf16_to_utf8 ((const gunichar2 *)file, -1);
 	if (!file_name) {
 		*image = NULL;
 		return InvalidParameter;
@@ -1120,28 +1203,28 @@ gdip_get_imageformat_from_codec_clsid (CLSID *encoderCLSID)
 {
 	GpStatus status;
 	UINT numEncoders, size, cnt;
-    	ImageCodecInfo *encoders, *encoder;
+	ImageCodecInfo *encoders, *encoder;
 
 	status = GdipGetImageEncodersSize (&numEncoders, &size);
 	if ((status != Ok) ||(numEncoders == 0))
 		return INVALID;
 
-    	encoders = GdipAlloc (size);
+	encoders = GdipAlloc (size);
 
 	status = GdipGetImageEncoders (numEncoders, size, encoders);
 	if (status != Ok)
 		return INVALID;
 
-    	for (cnt = 0, encoder = encoders; cnt < numEncoders; cnt++, encoder++) {
-	       	if (memcmp (&encoder->Clsid, encoderCLSID, sizeof (GUID)) == 0) {
-			ImageFormat ifmt = gdip_image_format_for_format_guid (&encoder->FormatID);
-            		GdipFree (encoders);
-			return ifmt;
-        	}
-    	}
+	for (cnt = 0, encoder = encoders; cnt < numEncoders; cnt++, encoder++) {
+		if (memcmp (&encoder->Clsid, encoderCLSID, sizeof (GUID)) == 0) {
+		ImageFormat ifmt = gdip_image_format_for_format_guid (&encoder->FormatID);
+				GdipFree (encoders);
+		return ifmt;
+		}
+	}
 
-    	GdipFree (encoders);
-    	return INVALID;
+	GdipFree (encoders);
+	return INVALID;
 }
 
 GpStatus WINGDIPAPI
@@ -1162,7 +1245,7 @@ GdipSaveImageToFile (GpImage *image, GDIPCONST WCHAR *file, GDIPCONST CLSID *enc
 	if (format == INVALID)
 		return UnknownImageFormat;
 	
-	file_name = (char *) ucs2_to_utf8 ((const gunichar2 *)file, -1);
+	file_name = (char *) utf16_to_utf8 ((const gunichar2 *)file, -1);
 	if (file_name == NULL)
 		return InvalidParameter;
 	
@@ -1247,8 +1330,11 @@ GdipGetImageDimension (GpImage *image, REAL *width, REAL *height)
 		break;
 	case ImageTypeMetafile: {
 		MetafileHeader *metaheader = gdip_get_metaheader(image);
-		*width = metaheader->Width * METAFILE_DIMENSION_FACTOR / metaheader->DpiX;
-		*height = metaheader->Height * METAFILE_DIMENSION_FACTOR / metaheader->DpiY;
+
+		// The width and height values are returned in 0.01 millimeter units.
+		*width = gdip_unit_conversion (UnitPixel, UnitMillimeter, metaheader->DpiX, gtMemoryBitmap, metaheader->Width) * 100;
+		*height = gdip_unit_conversion (UnitPixel, UnitMillimeter, metaheader->DpiY, gtMemoryBitmap, metaheader->Height) * 100;
+
 		break;
 	}
 	default:
@@ -1369,7 +1455,7 @@ GdipGetImageRawFormat (GpImage *image, GUID *format)
 		return InvalidParameter;
 	
 	switch (image->image_format) {
-       	case BMP:
+	case BMP:
 		memcpy (format, &gdip_bmp_image_format_guid, sizeof (GUID));
 		break;
 	case TIF:
@@ -1401,7 +1487,7 @@ GdipGetImageRawFormat (GpImage *image, GUID *format)
 		break;
 	default:
 		return InvalidParameter;
-    	}
+		}
 	return Ok;
 }
 
@@ -1449,11 +1535,11 @@ GdipImageGetFrameDimensionsList (GpImage *image, GUID *dimensionGUID, UINT count
 	if (!image || !dimensionGUID)
 		return InvalidParameter;
 
-	if (count < 1)
-		return InvalidParameter;
-
 	switch (image->type) {
 	case ImageTypeBitmap: {
+		if (count < 1 || count > image->num_of_frames)
+			return Win32Error;
+
 		int i;
 		int countReturn = image->num_of_frames;
 	
@@ -1465,8 +1551,9 @@ GdipImageGetFrameDimensionsList (GpImage *image, GUID *dimensionGUID, UINT count
 		break;
 	}
 	case ImageTypeMetafile:
-		if (count > 1)
+		if (count != 1)
 			return InvalidParameter;
+
 		dimensionGUID[0] = gdip_image_frameDimension_page_guid;
 		break;
 	default:
@@ -1478,11 +1565,14 @@ GdipImageGetFrameDimensionsList (GpImage *image, GUID *dimensionGUID, UINT count
 GpStatus WINGDIPAPI
 GdipImageGetFrameCount (GpImage *image, GDIPCONST GUID *dimensionGUID, UINT* count)
 {
-	if (!image || !dimensionGUID || !count)
+	if (!image)
 		return InvalidParameter;
 
 	switch (image->type) {
 	case ImageTypeBitmap: {
+		if (!dimensionGUID || !count)
+			return Win32Error;
+
 		int i;
 		for (i = 0; i < image->num_of_frames; i++){
 			if (memcmp (dimensionGUID, &(image->frames[i].frame_dimension), sizeof (CLSID)) == 0) {
@@ -1490,8 +1580,13 @@ GdipImageGetFrameCount (GpImage *image, GDIPCONST GUID *dimensionGUID, UINT* cou
 				return Ok;
 			}
 		}
+
+		return Win32Error;
 	}
 	case ImageTypeMetafile:
+		if (!count)
+			return InvalidParameter;
+
 		/* note: an empty GUID also returns 1 */
 		*count = 1;
 		return Ok;
@@ -1602,7 +1697,11 @@ gdip_rotate_orthogonal_flip_x (GpImage *image, int angle, BOOL flip_x)
 		}
 	}
 
-	rotated = GdipAlloc (target_height * target_stride);
+	unsigned long long int size = (unsigned long long int)target_height * target_stride;
+	if (size > G_MAXINT32)
+		return OutOfMemory;
+
+	rotated = GdipAlloc (size);
 
 	if (rotated == NULL) {
 		return OutOfMemory;
@@ -1612,14 +1711,14 @@ gdip_rotate_orthogonal_flip_x (GpImage *image, int angle, BOOL flip_x)
 	target = initial_target_offset + rotated;
 
 	for (y = 0; y < source_height;
-             y++,
-             source += source_interscan_delta,
-             target += target_interscan_delta) {
+		y++,
+		source += source_interscan_delta,
+		target += target_interscan_delta) {
 		for (x = 0; x < source_width;
-                     x++,
-                     source += source_pixel_delta,
-                     target += target_pixel_delta) {
-				copy_pixel (source, target, pixel_size);
+			x++,
+			source += source_pixel_delta,
+			target += target_pixel_delta) {
+			copy_pixel (source, target, pixel_size);
 		}
 	}
 
@@ -1680,7 +1779,11 @@ gdip_rotate_flip_packed_indexed (GpImage *image, PixelFormat pixel_format, int a
 		return gdip_flip_y(image);
 	}
 
-	rotated = GdipAlloc (target_height * target_stride);
+	unsigned long long int size = (unsigned long long int)target_height * target_stride;
+	if (size > G_MAXINT32)
+		return OutOfMemory;
+
+	rotated = GdipAlloc (size);
 	if (rotated == NULL) {
 		return OutOfMemory;
 	}
@@ -1835,7 +1938,7 @@ GdipImageRotateFlip (GpImage *image, RotateFlipType type)
 	angle = flip_x = 0;
 
 	switch (type) {
-                case RotateNoneFlipNone:/* equivalent to Rotate180FlipXY */
+		case RotateNoneFlipNone:/* equivalent to Rotate180FlipXY */
 			return Ok;
 
 		case Rotate90FlipNone:	/* equivalent to Rotate270FlipXY */
@@ -1868,7 +1971,7 @@ GdipImageRotateFlip (GpImage *image, RotateFlipType type)
 			break;
 
 		default:
-			return NotImplemented;
+			return InvalidParameter;
 	}	
 	
 	if (gdip_is_an_indexed_pixelformat (image->active_bitmap->pixel_format) && (gdip_get_pixel_format_depth (image->active_bitmap->pixel_format) < 8)) {
@@ -1881,36 +1984,33 @@ GdipImageRotateFlip (GpImage *image, RotateFlipType type)
 GpStatus WINGDIPAPI 
 GdipGetImagePalette (GpImage *image, ColorPalette *palette, INT size)
 {
-	int palette_entries;
 	int bytes_needed;
 
-        if (!image || !palette)
+	if (!image || !palette)
 		return InvalidParameter;
 
 	/* GDI+ doesn't support this for metafiles */
 	if (image->type != ImageTypeBitmap)
 		return NotImplemented;
 
-	if (!image->active_bitmap->palette) {
-		/* ColorPalette definition includes one ARGB member (ARGB Entries[1]). In reality there are 
-		 * Count entries allocated, where Count can be 0 (and required to substract the ARGB size)
-		 */
-		/* coverity[buffer_alloc] */
-		image->active_bitmap->palette = (ColorPalette*) GdipAlloc (sizeof(ColorPalette) - sizeof(ARGB));
-		image->active_bitmap->palette->Flags = 0;
-		image->active_bitmap->palette->Count = 0;
+	if (!image->active_bitmap->palette || image->active_bitmap->palette->Count == 0) {
+		if (size >= 0 && size < sizeof(ColorPalette))
+			return InvalidParameter;
+
+		palette->Count = 0;
+		if (image->active_bitmap->palette) {
+			palette->Flags = image->active_bitmap->palette->Flags;
+		}
+		else {
+			palette->Flags = 0;
+		}
+
+		return Ok;
 	}
 
-	palette_entries = image->active_bitmap->palette->Count;
-
-	if (image->active_bitmap->pixel_format == PixelFormat4bppIndexed)
-		palette_entries = 16;
-
-	bytes_needed = (palette_entries - 1) * sizeof(ARGB) + sizeof(ColorPalette);
-
-	if (bytes_needed > size) {
+	GdipGetImagePaletteSize (image, &bytes_needed);
+	if (bytes_needed != size)
 		return InvalidParameter;
-	}
 
 	memcpy(palette, image->active_bitmap->palette, bytes_needed);
 	return Ok;
@@ -1920,8 +2020,9 @@ GpStatus WINGDIPAPI
 GdipSetImagePalette (GpImage *image, GDIPCONST ColorPalette *palette)
 {
 	int size;
+	ColorPalette *result;
 
-        if (!image || !palette)
+	if (!image || !palette || palette->Count == 0 || palette->Count > 256)
 		return InvalidParameter;
 
 	/* GDI+ doesn't support this for metafiles */
@@ -1930,9 +2031,13 @@ GdipSetImagePalette (GpImage *image, GDIPCONST ColorPalette *palette)
 
 	size = (palette->Count - 1) * sizeof(ARGB) + sizeof(ColorPalette);
 	if (!image->active_bitmap->palette || (palette->Count != image->active_bitmap->palette->Count)) {
+		result = (ColorPalette *) GdipAlloc (size);
+		if (!result)
+			return OutOfMemory;
+
 		if (image->active_bitmap->palette)
 			GdipFree (image->active_bitmap->palette);
-		image->active_bitmap->palette = (ColorPalette*) GdipAlloc (size);
+		image->active_bitmap->palette = result;
 	}
 
 	memcpy (image->active_bitmap->palette, palette, size);
@@ -1942,21 +2047,22 @@ GdipSetImagePalette (GpImage *image, GDIPCONST ColorPalette *palette)
 GpStatus WINGDIPAPI 
 GdipGetImagePaletteSize (GpImage *image, INT* size)
 {
-        int palette_entries;
+	int palette_entries;
 
-        if (!image || !size)
-                return InvalidParameter;
+	if (!image || !size)
+		return InvalidParameter;
 
 	/* GDI+ doesn't support this for metafiles */
 	if (image->type != ImageTypeBitmap)
 		return GenericError;
 
-        palette_entries = (image->active_bitmap->palette) ? image->active_bitmap->palette->Count : 0;
+	palette_entries = (image->active_bitmap->palette) ? image->active_bitmap->palette->Count : 0;
 
-        if (image->active_bitmap->pixel_format == PixelFormat4bppIndexed)
-                palette_entries = 16;
+	if (palette_entries == 0)
+		*size = sizeof(ColorPalette);
+	else
+		*size = (palette_entries - 1) * sizeof(ARGB) + sizeof(ColorPalette);
 
-	*size = (palette_entries - 1) * sizeof(ARGB) + sizeof(ColorPalette);
 	return Ok;
 }
 
@@ -2026,7 +2132,6 @@ GpStatus WINGDIPAPI
 GdipGetPropertyItem (GpImage *image, PROPID propID, UINT size, PropertyItem *buffer)
 {
 	int	index;
-	int	size_needed;
 
 	if (!image || !buffer)
 		return InvalidParameter;
@@ -2100,9 +2205,12 @@ GdipSetPropertyItem(GpImage *image, GDIPCONST PropertyItem *item)
 		return NotImplemented;
 
 	switch(image->image_format) {
+		case BMP:
 		case TIF:
+		case GIF:
 		case JPEG:
-		case PNG: {
+		case PNG:
+		case ICON: {
 			break;
 		}
 
@@ -2156,14 +2264,14 @@ GdipCloneImage (GpImage *image, GpImage **cloneImage)
 
 	switch (image->type){
 	case ImageTypeBitmap:
-               	gdip_bitmap_clone (image, cloneImage);
+		gdip_bitmap_clone (image, cloneImage);
 		gdip_bitmap_setactive(*cloneImage, NULL, 0);
 		return Ok;
 
 	case ImageTypeMetafile:
 		return gdip_metafile_clone ((GpMetafile*)image, (GpMetafile**)cloneImage);
 
-       	case ImageTypeUnknown:
+	case ImageTypeUnknown:
 	default:
 		return Ok;
 	}
@@ -2245,10 +2353,10 @@ GpStatus WINGDIPAPI
 GdipLoadImageFromDelegate_linux (GetHeaderDelegate getHeaderFunc,
 								 GetBytesDelegate getBytesFunc,
 								 PutBytesDelegate putBytesFunc,
-                                 SeekDelegate seekFunc,
+								 SeekDelegate seekFunc,
 								 CloseDelegate closeFunc,
 								 SizeDelegate sizeFunc,
-                                 GpImage **image)
+								 GpImage **image)
 {
 	GpImage *result = 0;
 	GpStatus status = 0;
@@ -2318,14 +2426,14 @@ GdipSaveImageToDelegate_linux (GpImage *image, GetBytesDelegate getBytesFunc, Pu
 	SeekDelegate seekFunc, CloseDelegate closeFunc, SizeDelegate sizeFunc, GDIPCONST CLSID *encoderCLSID,
 	GDIPCONST EncoderParameters *params)
 {
-    	if (!image || !encoderCLSID || (image->type != ImageTypeBitmap))
-        	return InvalidParameter;
+	if (!image || !encoderCLSID || (image->type != ImageTypeBitmap))
+		return InvalidParameter;
 
 	switch (gdip_get_imageformat_from_codec_clsid ((CLSID *)encoderCLSID)) {
 	case ICON:
 	case BMP:
 		return gdip_save_bmp_image_to_stream_delegate (putBytesFunc, image);
-       	case PNG:
+	case PNG:
 		return gdip_save_png_image_to_stream_delegate (putBytesFunc, image, params);
 	case JPEG:
 		return gdip_save_jpeg_image_to_stream_delegate (putBytesFunc, image, params);
@@ -2364,11 +2472,6 @@ initCodecList (void)
 	memcpy (epos, gdip_getcodecinfo_bmp (), sizeof (ImageCodecInfo));
 	epos += sizeof (ImageCodecInfo);
 	g_encoders++;
-
-	/* ICO codec (decoder-only) - built-in */
-	memcpy (dpos, gdip_getcodecinfo_ico (), sizeof (ImageCodecInfo));
-	dpos += sizeof (ImageCodecInfo);
-	g_decoders++;
 	
 	/* JPEG codec (encoder+decoder) */
 	if (gdip_getcodecinfo_jpeg ()) {
@@ -2389,6 +2492,20 @@ initCodecList (void)
 		epos += sizeof (ImageCodecInfo);
 		g_encoders++;
 	}
+
+	/* EMF codec (decoder-only) */
+	if (gdip_getcodecinfo_emf ()) {
+		memcpy (dpos, gdip_getcodecinfo_emf (), sizeof (ImageCodecInfo));
+		dpos += sizeof (ImageCodecInfo);
+		g_decoders++;
+	}
+
+	/* WMF codec (decoder-only) */
+	if (gdip_getcodecinfo_wmf ()) {
+		memcpy (dpos, gdip_getcodecinfo_wmf (), sizeof (ImageCodecInfo));
+		dpos += sizeof (ImageCodecInfo);
+		g_decoders++;
+	}
 	
 	/* TIFF codec (encoder+decoder) */
 	if (gdip_getcodecinfo_tiff ()) {
@@ -2410,19 +2527,10 @@ initCodecList (void)
 		g_encoders++;
 	}
 
-	/* WMF codec (decoder-only) */
-	if (gdip_getcodecinfo_wmf ()) {
-		memcpy (dpos, gdip_getcodecinfo_wmf (), sizeof (ImageCodecInfo));
-		dpos += sizeof (ImageCodecInfo);
-		g_decoders++;
-	}
-
-	/* EMF codec (decoder-only) */
-	if (gdip_getcodecinfo_emf ()) {
-		memcpy (dpos, gdip_getcodecinfo_emf (), sizeof (ImageCodecInfo));
-		dpos += sizeof (ImageCodecInfo);
-		g_decoders++;
-	}
+	/* ICO codec (decoder-only) - built-in */
+	memcpy (dpos, gdip_getcodecinfo_ico (), sizeof (ImageCodecInfo));
+	dpos += sizeof (ImageCodecInfo);
+	g_decoders++;
 
 	return Ok;
 }
@@ -2510,6 +2618,10 @@ GdipGetAllPropertyItems (GpImage *image, UINT totalBufferSize, UINT numPropertie
 		return InvalidParameter;
 	}
 
+	if (numProperties == 0) {
+		return GenericError;
+	}
+
 	ptr = (BYTE*)allItems;
 	ptr += totalBufferSize;
 
@@ -2533,16 +2645,37 @@ GdipGetEncoderParameterListSize (GpImage *image, GDIPCONST CLSID *clsidEncoder,
 {
 	ImageFormat fmt;
 
-	if (!image || !clsidEncoder || !size)
+	if (!image || !clsidEncoder)
 		return InvalidParameter;
 
 	fmt = gdip_get_imageformat_from_codec_clsid ((CLSID *) clsidEncoder);
 
 	switch (fmt) {
-		case JPEG:
-			*size = gdip_get_encoder_parameter_list_size_jpeg ();
-			return Ok;
 		case TIF:
+			if (!size)
+				return InvalidParameter;
+
+			*size = sizeof (TiffEncoderParameters);
+			return Ok;
+		case GIF:
+			if (!size)
+				return InvalidParameter;
+
+			*size = sizeof (GifEncoderParameters);
+			return Ok;
+		case PNG:
+			if (!size)
+				return InvalidParameter;
+
+			*size = sizeof (PngEncoderParameters);
+			return Ok;
+		case JPEG:
+			if (!size)
+				return InvalidParameter;
+
+			*size = sizeof (JpegEncoderParameters);
+			return Ok;
+		case BMP:
 			*size = 0;
 			return NotImplemented;
 		default:
@@ -2557,15 +2690,21 @@ GdipGetEncoderParameterList (GpImage *image, GDIPCONST CLSID *clsidEncoder, UINT
 {
 	ImageFormat fmt;
 
-	if (!image || !clsidEncoder || !buffer)
+	if (!image || !clsidEncoder)
 		return InvalidParameter;
 
 	fmt = gdip_get_imageformat_from_codec_clsid ((CLSID *) clsidEncoder);
 
 	switch (fmt) {
+		case TIF:
+			return gdip_fill_encoder_parameter_list_tiff (buffer, size);
+		case GIF:
+			return gdip_fill_encoder_parameter_list_gif (buffer, size);
+		case PNG:
+			return gdip_fill_encoder_parameter_list_png (buffer, size);
 		case JPEG:
 			return gdip_fill_encoder_parameter_list_jpeg (buffer, size);
-		case TIF:
+		case BMP:
 			return NotImplemented;
 		default:
 			break;
@@ -2577,8 +2716,50 @@ GdipGetEncoderParameterList (GpImage *image, GDIPCONST CLSID *clsidEncoder, UINT
 GpStatus WINGDIPAPI
 GdipGetImageThumbnail (GpImage *image, UINT thumbWidth, UINT thumbHeight, GpImage **thumbImage, GetThumbnailImageAbort callback, VOID *callbackData)
 {
-	/* We don't really need to implement this; we've got it handled in System.Drawing directly */
-	return NotImplemented;
+	GpStatus status;
+	PixelFormat format;
+	GpImage *result;
+	GpGraphics *graphics;
+
+	if (!image || !thumbImage)
+		return InvalidParameter;
+
+	if (!thumbWidth && !thumbHeight)
+		thumbWidth = thumbHeight = 120;
+	else if (!thumbWidth || !thumbHeight)
+		return OutOfMemory;
+
+	switch (image->type) {
+	case ImageTypeBitmap:
+		format = PixelFormat32bppPARGB;
+		break;
+	case ImageTypeMetafile:
+		format = PixelFormat32bppARGB;
+		break;
+	default:
+		return InvalidParameter;
+	}
+
+	status = GdipCreateBitmapFromScan0 (thumbWidth, thumbHeight, 0, format, NULL, (GpBitmap **) &result);
+	if (status != Ok)
+		return status;
+
+	status = GdipGetImageGraphicsContext (result, &graphics);
+	if (status != Ok) {
+		GdipDisposeImage (result);
+		return status;
+	}
+
+	status = GdipDrawImageRectI (graphics, image, 0, 0, thumbWidth, thumbHeight);
+	if (status != Ok) {
+		GdipDisposeImage (result);
+		GdipDeleteGraphics (graphics);
+		return status;
+	}
+
+	GdipDeleteGraphics (graphics);
+	*thumbImage = result;
+	return Ok;
 }
 
 /* coverity[+alloc : arg-*1] */
@@ -2621,3 +2802,12 @@ GdipSaveAddImage (GpImage *image, GpImage *imageNew, GDIPCONST EncoderParameters
 		return InvalidParameter;
 	return NotImplemented;
 }
+
+GpStatus WINGDIPAPI
+GdipImageForceValidation (GpImage *image)
+{
+	if (!image)
+		return InvalidParameter;
+
+	return Ok;
+}
