@@ -27,9 +27,11 @@
  */
 
 #include "texturebrush-private.h"
+#include "general-private.h"
 #include "graphics-private.h"
 #include "bitmap-private.h"
 #include "matrix-private.h"
+#include "metafile-private.h"
 
 static GpStatus gdip_texture_setup (GpGraphics *graphics, GpBrush *brush);
 static GpStatus gdip_texture_clone (GpBrush *brush, GpBrush **clonedBrush);
@@ -48,8 +50,8 @@ static BrushClass vtable = { BrushTypeTextureFill,
 static cairo_content_t
 from_cairoformat_to_content (cairo_format_t format)
 {
-    	switch (format) {
-    	case CAIRO_FORMAT_RGB24:
+	switch (format) {
+	case CAIRO_FORMAT_RGB24:
 		return CAIRO_CONTENT_COLOR;
 	case CAIRO_FORMAT_A8:
 		return CAIRO_CONTENT_ALPHA;
@@ -75,11 +77,11 @@ gdip_texture_init (GpTexture *texture)
 static GpTexture*
 gdip_texture_new (void)
 {
-        GpTexture *result = (GpTexture *) GdipAlloc (sizeof (GpTexture));
+	GpTexture *result = (GpTexture *) GdipAlloc (sizeof (GpTexture));
 	if (result)
-	        gdip_texture_init (result);
+		gdip_texture_init (result);
 
-        return result;
+	return result;
 }
 
 /* 
@@ -104,7 +106,7 @@ draw_tile_texture (cairo_t *ct, GpBitmap *bitmap, GpTexture *brush)
 	if (gdip_bitmap_format_needs_premultiplication (bitmap)) {
 		premul = gdip_bitmap_get_premultiplied_scan0 (bitmap);
 		if (premul) {
-			BitmapData *data = bitmap->active_bitmap;
+			ActiveBitmapData *data = bitmap->active_bitmap;
 			original = cairo_image_surface_create_for_data (premul, CAIRO_FORMAT_ARGB32, 
 				data->width, data->height, data->stride);
 		}
@@ -176,7 +178,7 @@ draw_tile_flipX_texture (cairo_t *ct, GpBitmap *bitmap, GpTexture *brush)
 	if (gdip_bitmap_format_needs_premultiplication (bitmap)) {
 		premul = gdip_bitmap_get_premultiplied_scan0 (bitmap);
 		if (premul) {
-			BitmapData *data = bitmap->active_bitmap;
+			ActiveBitmapData *data = bitmap->active_bitmap;
 			original = cairo_image_surface_create_for_data (premul, CAIRO_FORMAT_ARGB32, 
 				data->width, data->height, data->stride);
 		}
@@ -262,7 +264,7 @@ draw_tile_flipY_texture (cairo_t *ct, GpBitmap *bitmap, GpTexture *brush)
 	if (gdip_bitmap_format_needs_premultiplication (bitmap)) {
 		premul = gdip_bitmap_get_premultiplied_scan0 (bitmap);
 		if (premul) {
-			BitmapData *data = bitmap->active_bitmap;
+			ActiveBitmapData *data = bitmap->active_bitmap;
 			original = cairo_image_surface_create_for_data (premul, CAIRO_FORMAT_ARGB32, 
 				data->width, data->height, data->stride);
 		}
@@ -352,7 +354,7 @@ draw_tile_flipXY_texture (cairo_t *ct, GpBitmap *bitmap, GpTexture *brush)
 	if (gdip_bitmap_format_needs_premultiplication (bitmap)) {
 		premul = gdip_bitmap_get_premultiplied_scan0 (bitmap);
 		if (premul) {
-			BitmapData *data = bitmap->active_bitmap;
+			ActiveBitmapData *data = bitmap->active_bitmap;
 			original = cairo_image_surface_create_for_data (premul, CAIRO_FORMAT_ARGB32, 
 				data->width, data->height, data->stride);
 		}
@@ -476,7 +478,7 @@ draw_clamp_texture (cairo_t *ct, GpBitmap *bitmap, GpTexture *brush)
 	/* texture surface to be created */
 	texture = cairo_surface_create_similar (original, from_cairoformat_to_content (bitmap->cairo_format),
 						rect->Width, rect->Height);
-	if (texture == NULL) {
+	if (!texture) {
 		cairo_pattern_destroy (pat);
 		return OutOfMemory;
 	}
@@ -528,9 +530,9 @@ gdip_texture_setup (GpGraphics *graphics, GpBrush *brush)
 		/* Unable to create a surface for the bitmap; it is an indexed image.
 		 * Instead, it will first be converted to 32-bit RGB. */
 		img = gdip_convert_indexed_to_rgb (img);
-		if (img == NULL) {
+		if (!img)
 			return OutOfMemory;
-		}
+
 		gdip_bitmap_ensure_surface (img);
 		dispose_bitmap = TRUE;
 	} else {
@@ -541,10 +543,9 @@ gdip_texture_setup (GpGraphics *graphics, GpBrush *brush)
 
 	/* We create the new pattern for brush, if the brush is changed
 	 * or if pattern has not been created yet. */
-	if (texture->base.changed || (texture->pattern == NULL)) {
-		if (texture->pattern != NULL) {
+	if (texture->base.changed || !texture->pattern) {
+		if (texture->pattern)
 			cairo_pattern_destroy (texture->pattern);
-		}
 
 		switch (texture->wrapMode) {
 			case WrapModeTile:
@@ -586,7 +587,7 @@ gdip_texture_setup (GpGraphics *graphics, GpBrush *brush)
 		if (!gdip_is_matrix_empty (&texture->matrix)) {
 			cairo_matrix_t product;
 
-		        gdip_cairo_matrix_copy (&product, &texture->matrix);
+			gdip_cairo_matrix_copy (&product, &texture->matrix);
 			cairo_matrix_invert (&product);
 			cairo_pattern_set_matrix (pattern, &product);
 		}
@@ -606,10 +607,9 @@ gdip_texture_clone (GpBrush *brush, GpBrush **clonedBrush)
 	if (!brush || !clonedBrush)
 		return InvalidParameter;
 
-	result = (GpTexture *) GdipAlloc (sizeof (GpTexture));
-	if (result == NULL) {
+	result = gdip_texture_new ();
+	if (!result)
 		return OutOfMemory;
-	}
 
 	texture = (GpTexture *) brush;
 	result->base = texture->base;
@@ -628,16 +628,14 @@ gdip_texture_clone (GpBrush *brush, GpBrush **clonedBrush)
 	result->image = NULL;
 	status = GdipCloneImage (texture->image, &result->image);
 	if (status != Ok) {
-		if (result->image)
-			GdipDisposeImage (result->image);
-		GdipFree (result);
-		result = NULL;
+		GdipDeleteBrush ((GpBrush *) result);
+		*clonedBrush = NULL;
+		return status;
 	} else {
 		cairo_surface_reference (result->image->surface);
 	}
 
 	*clonedBrush = (GpBrush *) result;
-
 	return status;
 }
 
@@ -662,37 +660,25 @@ gdip_texture_destroy (GpBrush *brush)
 	return Ok;
 }
 
-/* coverity[+alloc : arg-*2] */
-GpStatus WINGDIPAPI
-GdipCreateTexture (GpImage *image, GpWrapMode wrapMode, GpTexture **texture)
+static GpStatus
+gdip_texture_create_from_cloned_image (GpImage *image, GpWrapMode wrapMode, GpTexture **texture)
 {
+	GpTexture *result;
 	cairo_surface_t *imageSurface = NULL;
-	GpTexture	*result;
-	GpStatus status;
-
-	if (!image || !texture)
-		return InvalidParameter;
-
-	if ((wrapMode < WrapModeTile) || (wrapMode > WrapModeClamp))
-		return OutOfMemory;
-
-	if (image->type != ImageTypeBitmap)
-		return NotImplemented;
 
 	result = gdip_texture_new ();
 	if (!result)
 		return OutOfMemory;
 
-	result->image = NULL;
-	status = GdipCloneImage (image, &result->image);
-	if (status != Ok)
-		goto failure;
+	result->image = image;
 
 	/* note: we must keep the scan0 alive, so we must use the cloned image (and not the original) see bug #80971 */
 	imageSurface = cairo_image_surface_create_for_data ((BYTE*)result->image->active_bitmap->scan0,
-		image->cairo_format, image->active_bitmap->width, image->active_bitmap->height, image->active_bitmap->stride);
-	if (!imageSurface)
-		goto failure;
+		result->image->cairo_format, result->image->active_bitmap->width, result->image->active_bitmap->height, result->image->active_bitmap->stride);
+	if (!imageSurface) {
+		GdipDeleteBrush ((GpBrush *) result);
+		return OutOfMemory;
+	}
 
 	result->wrapMode = wrapMode;
 	if (result->image->surface)
@@ -700,99 +686,149 @@ GdipCreateTexture (GpImage *image, GpWrapMode wrapMode, GpTexture **texture)
 	result->image->surface = imageSurface;
 	result->rectangle.X = 0;
 	result->rectangle.Y = 0;
-	result->rectangle.Width = image->active_bitmap->width;
-	result->rectangle.Height = image->active_bitmap->height;
+	result->rectangle.Width = result->image->active_bitmap->width;
+	result->rectangle.Height = result->image->active_bitmap->height;
 
 	*texture = result;
 	return Ok;
-
-failure:
-	if (result->image)
-		GdipDisposeImage (result->image);
-	if (imageSurface)
-		cairo_surface_destroy (imageSurface);
-	GdipFree (result);
-	*texture = NULL;
-	return status;
 }
 
-/* coverity[+alloc : arg-*6] */
+/* coverity[+alloc : arg-*2] */
 GpStatus WINGDIPAPI
-GdipCreateTexture2 (GpImage *image, GpWrapMode wrapMode, float x, float y, float width, float height, GpTexture **texture)
+GdipCreateTexture (GpImage *image, GpWrapMode wrapmode, GpTexture **texture)
 {
-	return GdipCreateTexture2I (image, wrapMode, (int) x, (int) y, (int) width, (int) height, texture);
-}
-
-/* coverity[+alloc : arg-*6] */
-GpStatus WINGDIPAPI
-GdipCreateTexture2I (GpImage *image, GpWrapMode wrapMode, int x, int y, int width, int height, GpTexture **texture)
-{
-	int bmpWidth;
-	int bmpHeight;
+	GpImage *textureImage;
 	GpStatus status;
-	GpImage *resized_image = NULL;
+
+	if (!gdiplusInitialized)
+		return GdiplusNotInitialized;
 
 	if (!image || !texture)
 		return InvalidParameter;
 
-	if (image->type != ImageTypeBitmap)
-		return NotImplemented;
-
-	bmpWidth = image->active_bitmap->width;
-	bmpHeight = image->active_bitmap->height;
-
-	/* MS behaves this way */
-	if ((x < 0) || (y < 0) || (width <= 0) || (height <= 0) || (bmpWidth < (x + width)) || (bmpHeight < (y + height)))
+	if (wrapmode > WrapModeClamp) {
+		*texture = NULL;
 		return OutOfMemory;
+	}
 
-	status = GdipCloneBitmapAreaI (x, y, width, height, image->active_bitmap->pixel_format, image, &resized_image);
-	if (status != Ok)
-		return status;
+	switch (image->type) {
+	case ImageTypeBitmap:
+		status = GdipCloneImage (image, &textureImage);
+		if (status != Ok)
+			return status;
+		break;
+	case ImageTypeMetafile:
+		status = gdip_get_bitmap_from_metafile ((GpMetafile *) image, 0, 0, &textureImage);
+		if (status != Ok)
+			return status;
+		break;
+	default:
+		return GenericError;
+	}
 
-	status = GdipCreateTexture (resized_image, wrapMode, texture);
-	GdipDisposeImage (resized_image);
-	return status;
+	return gdip_texture_create_from_cloned_image (textureImage, wrapmode, texture);
 }
 
 /* coverity[+alloc : arg-*6] */
 GpStatus WINGDIPAPI
-GdipCreateTextureIA (GpImage *image, GpImageAttributes *imageAttributes, float x, float y, float width, float height, GpTexture **texture)
+GdipCreateTexture2 (GpImage *image, GpWrapMode wrapmode, REAL x, REAL y, REAL width, REAL height, GpTexture **texture)
 {
-	/* FIXME MonoTODO: Make use of ImageAttributes parameter when
-	 * ImageAttributes is implemented */
-	GpWrapMode mode = imageAttributes ? WrapModeClamp : WrapModeTile;
-	return GdipCreateTexture2 (image, mode, x, y, width, height, texture);
+	if (!gdiplusInitialized)
+		return GdiplusNotInitialized;
+
+	return GdipCreateTexture2I (image, wrapmode, (INT) x, (INT) y, (INT) width, (INT) height, texture);
 }
 
 /* coverity[+alloc : arg-*6] */
 GpStatus WINGDIPAPI
-GdipCreateTextureIAI (GpImage *image, GpImageAttributes *imageAttributes, int x, int y, int width, int height, GpTexture **texture)
+GdipCreateTexture2I (GpImage *image, GpWrapMode wrapmode, INT x, INT y, INT width, INT height, GpTexture **texture)
 {
+	GpImage *textureImage;
+	GpStatus status;
+
+	if (!gdiplusInitialized)
+		return GdiplusNotInitialized;
+
+	if (!image || !texture)
+		return InvalidParameter;
+
+	if (wrapmode > WrapModeClamp) {
+		*texture = NULL;
+		return OutOfMemory;
+	}
+
+	switch (image->type) {
+	case ImageTypeBitmap: {
+		INT bmpWidth = image->active_bitmap->width;
+		INT bmpHeight = image->active_bitmap->height;
+		if ((x < 0) || (y < 0) || (width <= 0) || (height <= 0) || (bmpWidth < (x + width)) || (bmpHeight < (y + height))) {
+			*texture = NULL;
+			return OutOfMemory;
+		}
+
+		status = GdipCloneBitmapAreaI (x, y, width, height, image->active_bitmap->pixel_format, image, &textureImage);
+		if (status != Ok)
+			return status;
+		break;
+	}
+	case ImageTypeMetafile:
+		status = gdip_get_bitmap_from_metafile ((GpMetafile *) image, width, height, &textureImage);
+		if (status != Ok)
+			return status;
+		break;
+	default:
+		return GenericError;
+	}
+	
+	return gdip_texture_create_from_cloned_image (textureImage, wrapmode, texture);
+}
+
+/* coverity[+alloc : arg-*6] */
+GpStatus WINGDIPAPI
+GdipCreateTextureIA (GpImage *image, GpImageAttributes *imageAttributes, REAL x, REAL y, REAL width, REAL height, GpTexture **texture)
+{
+	if (!gdiplusInitialized)
+		return GdiplusNotInitialized;
+
+	return GdipCreateTextureIAI (image, imageAttributes, (INT) x, (INT) y, (INT) width, (INT) height, texture);
+}
+
+/* coverity[+alloc : arg-*6] */
+GpStatus WINGDIPAPI
+GdipCreateTextureIAI (GpImage *image, GpImageAttributes *imageAttributes, INT x, INT y, INT width, INT height, GpTexture **texture)
+{
+	if (!gdiplusInitialized)
+		return GdiplusNotInitialized;
+
 	/* FIXME MonoTODO: Make use of ImageAttributes parameter when
 	 * ImageAttributes is implemented */
-	GpWrapMode mode = imageAttributes ? WrapModeClamp : WrapModeTile;
+	GpWrapMode mode = imageAttributes ? imageAttributes->wrapmode : WrapModeTile;
 	return GdipCreateTexture2I (image, mode, x, y, width, height, texture);
 }
 
 GpStatus WINGDIPAPI
 GdipGetTextureTransform (GpTexture *texture, GpMatrix *matrix)
 {
-	if ((texture == NULL) || (matrix == NULL)) {
+	if (!texture || !matrix)
 		return InvalidParameter;
-	}
 
-	gdip_cairo_matrix_copy(matrix, &texture->matrix);
+	gdip_cairo_matrix_copy (matrix, &texture->matrix);
 	return Ok;
 }
 
 GpStatus WINGDIPAPI
 GdipSetTextureTransform (GpTexture *texture, GDIPCONST GpMatrix *matrix)
 {
-	if ((texture == NULL) || (matrix == NULL)) {
-		return InvalidParameter;
-	}
+	BOOL invertible;
 
-	gdip_cairo_matrix_copy(&texture->matrix, matrix);
+	if (!texture || !matrix)
+		return InvalidParameter;
+		
+	GdipIsMatrixInvertible ((GpMatrix *) matrix, &invertible);
+	if (!invertible)
+		return InvalidParameter;
+
+	gdip_cairo_matrix_copy (&texture->matrix, matrix);
 	texture->base.changed = TRUE;
 
 	return Ok;
@@ -801,9 +837,8 @@ GdipSetTextureTransform (GpTexture *texture, GDIPCONST GpMatrix *matrix)
 GpStatus WINGDIPAPI
 GdipResetTextureTransform (GpTexture *texture)
 {
-	if (texture == NULL) {
+	if (!texture)
 		return InvalidParameter;
-	}
 
 	cairo_matrix_init_identity (&texture->matrix);
 	texture->base.changed = TRUE;
@@ -813,83 +848,84 @@ GdipResetTextureTransform (GpTexture *texture)
 GpStatus WINGDIPAPI
 GdipMultiplyTextureTransform (GpTexture *texture, GpMatrix *matrix, GpMatrixOrder order)
 {
-	GpStatus status;
-	BOOL invertible = FALSE;
-	cairo_matrix_t mat;
+	BOOL invertible;
 
-	if ((texture == NULL) || (matrix == NULL)) {
+	if (!texture)
 		return InvalidParameter;
-	}
+	
+	if (!matrix)
+		return Ok;
 
 	/* the matrix MUST be invertible to be used */
-	status = GdipIsMatrixInvertible ((GpMatrix*) matrix, &invertible);
-	if (!invertible || (status != Ok))
+	GdipIsMatrixInvertible (matrix, &invertible);
+	if (!invertible)
 		return InvalidParameter;
 
 	if (order == MatrixOrderPrepend)
-		cairo_matrix_multiply (&mat, matrix, &texture->matrix);
-	else if (order == MatrixOrderAppend)
-		cairo_matrix_multiply (&mat, &texture->matrix, matrix);
+		cairo_matrix_multiply (&texture->matrix, matrix, &texture->matrix);
+	else
+		cairo_matrix_multiply (&texture->matrix, &texture->matrix, matrix);
 
-	gdip_cairo_matrix_copy (&texture->matrix, &mat);
 	texture->base.changed = TRUE;
-	return status;
+	return Ok;
 }
 
 GpStatus WINGDIPAPI
-GdipTranslateTextureTransform (GpTexture *texture, float dx, float dy, GpMatrixOrder order)
+GdipTranslateTextureTransform (GpTexture *texture, REAL dx, REAL dy, GpMatrixOrder order)
 {
 	GpStatus status;
 
-	if (texture == NULL) {
+	if (!texture)
 		return InvalidParameter;
-	}
 
 	status = GdipTranslateMatrix (&texture->matrix, dx, dy, order);
-	if (status == Ok)
-		texture->base.changed = TRUE;
-	return status;
+	if (status != Ok)
+		return status;
+	
+	texture->base.changed = TRUE;
+	return Ok;
 }
 
 GpStatus WINGDIPAPI
-GdipScaleTextureTransform (GpTexture *texture, float sx, float sy, GpMatrixOrder order)
+GdipScaleTextureTransform (GpTexture *texture, REAL sx, REAL sy, GpMatrixOrder order)
 {
 	GpStatus status;
 
-	if (texture == NULL) {
+	if (!texture)
 		return InvalidParameter;
-	}
 
 	status = GdipScaleMatrix (&texture->matrix, sx, sy, order);
-	if (status == Ok)
-		texture->base.changed = TRUE;
-	return status;
+	if (status != Ok)
+		return status;
+
+	texture->base.changed = TRUE;
+	return Ok;
 }
 
-/* MonoTODO - FIXME - this hack affect the public transform and doesn't work with MultiplyTransform */
 GpStatus WINGDIPAPI
-GdipRotateTextureTransform (GpTexture *texture, float angle, GpMatrixOrder order)
+GdipRotateTextureTransform (GpTexture *texture, REAL angle, GpMatrixOrder order)
 {
 	GpStatus status;
 
-	if (texture == NULL)
+	if (!texture)
 		return InvalidParameter;
 
 	status = GdipRotateMatrix (&texture->matrix, angle, order);
-	if (status == Ok)
-		texture->base.changed = TRUE;
+	if (status != Ok)
+		return status;
 
-	return status;
+	texture->base.changed = TRUE;
+	return Ok;
 }
 
 GpStatus WINGDIPAPI
 GdipSetTextureWrapMode (GpTexture *texture, GpWrapMode wrapMode)
 {
-	if (texture == NULL)
+	if (!texture)
 		return InvalidParameter;
 
 	/* ignore invalid GpWrapMode value */
-	if ((wrapMode < WrapModeTile) || (wrapMode > WrapModeClamp))
+	if (wrapMode > WrapModeClamp)
 		return Ok;
 
 	texture->wrapMode = wrapMode;
@@ -912,6 +948,9 @@ GdipGetTextureWrapMode (GpTexture *texture, GpWrapMode *wrapMode)
 GpStatus WINGDIPAPI
 GdipGetTextureImage (GpTexture *texture, GpImage **image)
 {
+	if (!gdiplusInitialized)
+		return GdiplusNotInitialized;
+
 	if (!texture || !image)
 		return InvalidParameter;
 
